package net.minecraft.src.railcraft.common.api;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import net.minecraft.src.Chunk;
import net.minecraft.src.ChunkCoordIntPair;
import net.minecraft.src.Entity;
import net.minecraft.src.MathHelper;
import net.minecraft.src.World;
import net.minecraft.src.forge.IChunkLoadHandler;
import net.minecraft.src.railcraft.Railcraft;

/**
 * The ChunkManager is used to register anchor points
 * that work to keeps the chunks around them loaded.
 * @author CovertJaguar <railcraft.wikispaces.com>
 */
public class ChunkManager
{

    /**
     * Default radius of chunks to keep loaded.
     */
    public static final byte DEFAULT_WINDOW_RADIUS = 0;
    /**
     * Max radius of chunks to keep loaded.
     */
    public static final byte MAX_WINDOW_RADIUS = 4;
    private static ChunkManager instance;
    private ChunkHandler handler;
    private World activeWorld;
    private Map<WorldCoordinate, Byte> anchors = new HashMap<WorldCoordinate, Byte>();
    private Map<Entity, Byte> entities = new HashMap<Entity, Byte>();

    /**
     * Returns an instance of ChunkManager
     * @return ChunkManager
     */
    public static ChunkManager getInstance()
    {
        if(instance == null) {
            instance = new ChunkManager();
        }
        return instance;
    }

    /**
     * Returns the IChunkLoadHandler.
     * @return IChunkLoadHandler
     */
    public IChunkLoadHandler getHandler()
    {
        if(handler == null) {
            handler = new ChunkHandler();
        }
        return handler;
    }

    private ChunkManager()
    {
    }

    private class ChunkHandler implements IChunkLoadHandler
    {

        private ChunkHandler()
        {
        }

        @Override
        public void addActiveChunks(World world, Set<ChunkCoordIntPair> chunkList)
        {
            if(Railcraft.gameIsNotHost()) {
                return;
            }

            for(Map.Entry<WorldCoordinate, Byte> anchor : anchors.entrySet()) {
                int x = anchor.getKey().x >> 4;
                int z = anchor.getKey().z >> 4;

                chunkList.addAll(getChunksAround(world, x, z, anchor.getValue()));
            }

            Iterator<Map.Entry<Entity, Byte>> it = entities.entrySet().iterator();
            while(it.hasNext()) {
                Map.Entry<Entity, Byte> anchor = it.next();
                Entity entity = anchor.getKey();
                if(entity.isDead) {
                    it.remove();
                    continue;
                }

                int x = MathHelper.floor_double(entity.posX) >> 4;
                int z = MathHelper.floor_double(entity.posZ) >> 4;

                chunkList.addAll(getChunksAround(world, x, z, anchor.getValue()));

            }
        }

        @Override
        public boolean canUnloadChunk(Chunk chunk)
        {
            if(Railcraft.gameIsNotHost()) {
                return true;
            }

            for(Map.Entry<WorldCoordinate, Byte> anchor : anchors.entrySet()) {
                return canUnloadChunk(chunk, anchor.getKey(), anchor.getValue());
            }

            Iterator<Map.Entry<Entity, Byte>> it = entities.entrySet().iterator();
            while(it.hasNext()) {
                Map.Entry<Entity, Byte> anchor = it.next();
                Entity entity = anchor.getKey();
                if(entity.isDead) {
                    it.remove();
                    continue;
                }
                if(!canUnloadChunk(chunk, entity, anchor.getValue())) {
                    return false;
                }
            }

            return true;
        }

        private boolean canUnloadChunk(Chunk chunk, WorldCoordinate block, int radius)
        {
            int x = block.x >> 4;
            int z = block.z >> 4;

            if(chunk.worldObj.worldProvider.worldType == block.dimension
                && Math.abs(x - chunk.xPosition) <= radius && Math.abs(z - chunk.zPosition) <= radius) {
//            System.out.println("Keeping Chunk Loaded <" + chunk.xPosition + ", " + chunk.zPosition + ">");
                return false;
            }
            return true;
        }

        private boolean canUnloadChunk(Chunk chunk, Entity entity, int radius)
        {
            int x = MathHelper.floor_double(entity.posX) >> 4;
            int z = MathHelper.floor_double(entity.posZ) >> 4;

            if(chunk.worldObj.worldProvider.worldType == entity.worldObj.worldProvider.worldType
                && Math.abs(x - chunk.xPosition) <= radius && Math.abs(z - chunk.zPosition) <= radius) {
                return false;
            }
            return true;
        }
    }

    private byte validateWindow(int radius)
    {
        if(radius < 0) {
            radius = 0;
        }
        return (byte)Math.min(MAX_WINDOW_RADIUS, radius);
    }

    /**
     * Returns a Set of ChunkCoordIntPair containing the chunks around point [x, z].
     * Coordinates are in chunk coordinates, not world coordinates.
     * @param world World
     * @param x Chunk x-Coord
     * @param z Chunk z-Coord
     * @param radius Distance from [x, z] to include.
     * @return A set of chunks.
     */
    protected Set<ChunkCoordIntPair> getChunksAround(World world, int x, int z, int radius)
    {
        radius = validateWindow(radius);
        Set<ChunkCoordIntPair> chunkList = new HashSet<ChunkCoordIntPair>();
        for(int xx = x - radius; xx <= x + radius; xx++) {
            for(int zz = z - radius; zz <= z + radius; zz++) {
                chunkList.add(world.getChunkFromBlockCoords(xx, zz).getChunkCoordIntPair());
            }
        }
        return chunkList;
    }

    /**
     * Loads the chunks around point [x. z], with radius. 
     * Coordinates are in chunk coordinates, not world coordinates.
     * @param world World
     * @param x Chunk x-Coord
     * @param z Chunk z-Coord
     * @param radius Distance from [x, z] to load.
     */
    public void loadChunksAround(World world, int x, int z, int radius)
    {
        for(ChunkCoordIntPair chunk : getChunksAround(world, x, z, radius)) {
            if(!world.getChunkProvider().chunkExists(x, z)) {
                world.getChunkProvider().loadChunk(chunk.chunkXPos, chunk.chunkZPos);
            }
        }
    }

    /**
     * Registers a world coordinate as an anchor point that will keep the chunk
     * that coordinate is in loaded.
     * @param world World
     * @param i World x-Coord
     * @param j World y-Coord
     * @param k World z-Coord
     */
    public void registerAnchor(World world, int i, int j, int k)
    {
        registerAnchor(world, i, j, k, DEFAULT_WINDOW_RADIUS);
    }

    /**
     * Registers a world coordinate as an anchor point with a radius (in chunks) 
     * that will keep the chunks around it loaded.
     * @param world World
     * @param i World x-Coord
     * @param j World y-Coord
     * @param k World z-Coord
     * @param radius Radius (in chunks) to keep loaded
     */
    public void registerAnchor(World world, int i, int j, int k, int radius)
    {
        anchors.put(new WorldCoordinate(world.worldProvider.worldType, i, j, k), Byte.valueOf(validateWindow(radius)));
    }

    /**
     * Unregisters an anchor point.
     * @param world World
     * @param i World x-Coord
     * @param j World y-Coord
     * @param k World z-Coord
     */
    public void unregisterAnchor(World world, int i, int j, int k)
    {
        anchors.remove(new WorldCoordinate(world.worldProvider.worldType, i, j, k));
    }

    /**
     * Registers an entity as an anchor point
     * that will keep the chunk it is in loaded.
     * @param entity The entity
     */
    public void registerAnchor(Entity entity)
    {
        registerAnchor(entity, DEFAULT_WINDOW_RADIUS);
    }

    /**
     * Registers an entity as an anchor point
     * that will keep the chunks around it loaded.
     * @param entity The entity
     * @param radius Radius (in chunks) to keep loaded
     */
    public void registerAnchor(Entity entity, int radius)
    {
        entities.put(entity, Byte.valueOf(validateWindow(radius)));
    }

    /**
     * Unregisters an Entity as an anchor point.
     * @param entity The Entity
     */
    public void unregisterAnchor(Entity entity)
    {
        entities.remove(entity);
    }

    /**
     * Used by the client to detect World changes
     * and reset the ChunkManager if the World changed.
     * Not used on the server.
     * @param world The world
     */
    public void verifyWorld(World world)
    {
        if(activeWorld != world) {
            anchors.clear();
            entities.clear();
            activeWorld = world;
        }
    }
}
